
from unittest import TestCase

from mapping_shortcuts.utils import parse_args, first


class FirstTestCase(TestCase):
    def test_ok(self):
        itr = (i for i in range(10) if not i % 3)
        self.assertEqual(first(itr), 0)
        self.assertEqual(first(itr), 3)
        self.assertEqual(first(itr), 6)
        self.assertEqual(first(itr), 9)
        self.assertEqual(first(itr), None)
        self.assertEqual(first(itr), None)


class ParseArgsTestCase(TestCase):
    def test_none(self):
        res = parse_args(['a', 'b', 'c'])
        self.assertEqual(res, {})

    def test_flags(self):
        res = parse_args(['-a', '-b', '-c'])
        self.assertEqual(res, {'-a': True, '-b': True, '-c': True})

    def test_multiflags(self):
        res = parse_args(['-abc'])
        self.assertEqual(res, {'-abc': True})

    def test_params(self):
        res = parse_args(['--a=123', '--long-name=somearg', '--empty='])
        self.assertEqual(res, {'--a': '123', '--long-name': 'somearg', '--empty': ''})




from typing import Iterator, TypeVar, Union, Type, Callable, Any

T = TypeVar('T')

ParsedArgsType = dict[str, Union[str, bool, list[Union[str, bool]]]]


def parse_args(args: list[str]) -> ParsedArgsType:
    res = {}  # type: ParsedArgsType

    def insert(data: ParsedArgsType, key: str, value: Union[str, bool]) -> ParsedArgsType:
        if key in data:
            data[key] = [res[key], value]  # type: ignore[list-item]
        else:
            data[key] = value
        return data

    for st in args:
        if st.startswith('-'):
            if '=' in st:
                key, *value = st.split('=')
                res = insert(res, key, '='.join(value))
            else:
                res = insert(res, st, True)
    return res


def first(itr: Iterator[T]) -> Union[T, None]:
    try:
        return next(itr)
    except StopIteration:
        return None


def get_arg_types(func: Callable[..., Any]) -> Iterator[Type[T]]:
    annonations = getattr(func, '__annotations__', None) or {}  # type: dict[str, Type[T]]
    annonations.pop('return', None)
    yield from annonations.values()




from unittest import TestCase

from pydantic import BaseModel, Field

from mapping_shortcuts.cli import cli_handler, run_args, help_handler
from mapping_shortcuts.cli.cli import CMD_HANDLERS


class RunCliTestCase(TestCase):
    def test_simple(self):
        CMD_HANDLERS.clear()
        cmd1 = False
        cmd2 = False

        @cli_handler('cmd1')
        def handler(args):
            nonlocal cmd1
            cmd1 = True

        @cli_handler('cmd2')
        def handler(args):
            nonlocal cmd2
            cmd2 = True

        assert len(CMD_HANDLERS) == 2
        run_args(['cmd1'])
        self.assertTrue(cmd1)
        self.assertFalse(cmd2)

    def test_empty_run(self):
        CMD_HANDLERS.clear()
        cmd1 = 0

        @cli_handler('help')
        def handler(args):
            nonlocal cmd1
            cmd1 += 1

        assert len(CMD_HANDLERS) == 1
        run_args([])
        run_args(['-?'])
        run_args(['-h'])
        run_args(['--help'])
        self.assertEqual(cmd1, 4)

    def test_run_help(self):
        CMD_HANDLERS.clear()

        cli_handler('help')(help_handler)

        assert len(CMD_HANDLERS) == 1
        run_args([])
        # should not raise exceptions

    def test_model(self):
        CMD_HANDLERS.clear()

        class ArgModel(BaseModel):
            param1: str = Field(alias='--param1')
            param2: str = Field(alias='-p2')

        cmd1 = False

        @cli_handler('cmd1', model=ArgModel)
        def handler(args: ArgModel):
            nonlocal cmd1
            self.assertTrue(args.param2)
            self.assertEqual(args.param1, '123')
            cmd1 = True

        assert len(CMD_HANDLERS) == 1
        try:
            run_args(['cmd1'])
            self.assertTrue(False, 'expected exception')
        except ValueError:
            ...

        run_args(['cmd1', '--param1=123', '-p2'])
        self.assertTrue(cmd1)

    def test_annotation(self):
        CMD_HANDLERS.clear()

        class ArgModel1(BaseModel):
            param: str = Field(alias='--param')

        class ArgModel2(BaseModel):
            param: str = Field(alias='--param')

        @cli_handler('cmd1', model=ArgModel1)
        def handler_1(args):
            ...

        @cli_handler('cmd2')
        def handler_2(args: ArgModel2):
            ...

        assert len(CMD_HANDLERS) == 2
        self.assertIs(CMD_HANDLERS['cmd1'].func, handler_1)
        self.assertIs(CMD_HANDLERS['cmd1'].model, ArgModel1)
        self.assertIs(CMD_HANDLERS['cmd2'].func, handler_2)
        self.assertIs(CMD_HANDLERS['cmd2'].model, ArgModel2)

    def test_list_args(self):
        CMD_HANDLERS.clear()

        class ArgModel(BaseModel):
            param: list[int] = Field(alias='--param')

        @cli_handler('cmd1')
        def handler_1(args: ArgModel):
            self.assertEqual(args.param, [123, 456])

        @cli_handler('cmd2')
        def handler_2(args: ArgModel):
            self.assertEqual(args.param, [123])

        assert len(CMD_HANDLERS) == 2
        self.assertIs(CMD_HANDLERS['cmd1'].func, handler_1)
        self.assertIs(CMD_HANDLERS['cmd2'].func, handler_2)
        run_args(['cmd1', '--param=123', '--param=456'])
        run_args(['cmd2', '--param=123'])



import sys
from typing import Callable, Type, Union, Optional, TypeVar

from pydantic import BaseModel  # pylint: disable=no-name-in-module

from mapping_shortcuts.utils import first, parse_args, get_arg_types, ParsedArgsType

from . import consts

CliHandler = Union[Callable[[BaseModel], None], Callable[[ParsedArgsType], None]]
ModelT = TypeVar('ModelT', bound=BaseModel)


class Cmd(BaseModel):
    cmd: str
    desc: str
    func: CliHandler
    model: Optional[Type[BaseModel]] = None

    def __hash__(self) -> int:
        return hash(self.cmd)


CMD_HANDLERS = {}  # type: dict[str, Cmd]


def cli_handler(
    cmd: str,
    *,
    desc: str = '',
    model: Optional[Type[BaseModel]] = None,
) -> Callable[[CliHandler], CliHandler]:
    def decor(func: CliHandler) -> CliHandler:
        nonlocal model
        if not model:
            possible_model = first(get_arg_types(func))  # type: Optional[Type[BaseModel]]
            if possible_model:
                if BaseModel in possible_model.__mro__:
                    model = possible_model
        if cmd in CMD_HANDLERS:
            raise RuntimeError(f'key duplication for "{cmd}"')
        CMD_HANDLERS[cmd] = Cmd(cmd=cmd, desc=desc, func=func, model=model)
        return func
    return decor


def validate_args(model: Type[ModelT], params: ParsedArgsType) -> ModelT:
    for key, props in model.schema()['properties'].items():
        if key in params:
            if props['type'] == 'array':
                if not isinstance(params[key], list):
                    params[key] = [params[key]]  # type: ignore[list-item]

    return model(**params)


def run_args(args: list[str]) -> None:
    raw_params = parse_args(args)

    if {'-h', '-?', '--help'} & set(raw_params):
        cmd = 'help'
    else:
        cmd = first(i for i in args if not i.startswith('-')) or 'help'
        if cmd not in CMD_HANDLERS:
            raise ValueError(f'unknown cmd: {cmd}')

    handler = CMD_HANDLERS[cmd]
    if handler.model:
        validated_args = validate_args(handler.model, raw_params)
        handler.func(validated_args)  # type: ignore[call-arg]
    else:
        handler.func(raw_params)  # type: ignore[call-arg]


def process_sysargv(
    args: Optional[list[str]] = None,
    help_msg_header: Optional[str] = None,
    help_msg_run_cmd: Optional[str] = None,
    help_msg_template: Optional[str] = None,
) -> None:
    args = args or sys.argv

    if help_msg_header:
        consts.HELP_MSG_TEMPLATE_HEADER = help_msg_header
    if help_msg_run_cmd:
        consts.HELP_MSG_RUN_CMD = help_msg_run_cmd
    if help_msg_template:
        consts.HELP_TEMPLATE = help_msg_template

    run_args(args[1:])




from typing import Union

from .cli import CMD_HANDLERS, cli_handler
from . import consts


@cli_handler('help', desc='see this cool msg again')
def help_handler(args: dict[str, Union[str, bool]]) -> None:
    rows = []
    for handler in CMD_HANDLERS.values():
        rows.append(f'\t{handler.cmd} - {handler.desc}')
        if handler.model:
            rows.extend(
                [
                    f'\t\t{key} - {value.get("description", "")}'
                    for key, value in handler.model.schema()['properties'].items()
                ]
            )

    txt = consts.HELP_TEMPLATE.format(
        header=consts.HELP_MSG_TEMPLATE_HEADER,
        run_cmd=consts.HELP_MSG_RUN_CMD,
        commands='\n'.join(rows),
    )
    print(txt)




from unittest import TestCase

from mapping_shortcuts.decors import create_collector
from mapping_shortcuts.decors import create_class_collector


class DecorTestCase(TestCase):

    def test_ok(self):
        decor, collection = create_collector()

        @decor('1')
        def f1():
            ...

        @decor('2')
        def f2():
            ...

        self.assertEqual(len(collection), 2)
        self.assertIn('1', collection)
        self.assertIn('2', collection)
        self.assertIs(collection['1'], f1)
        self.assertIs(collection['2'], f2)

    def test_duplicate_raise(self):
        decor, collection = create_collector()

        @decor('1')
        def f1():
            ...

        try:
            @decor('1')
            def f2():
                ...
        except ValueError as ex:
            self.assertTrue(str(ex).startswith('Duplication for key'))
        else:
            self.assertTrue(False, 'expected exception here')

        self.assertEqual(len(collection), 1)
        self.assertIn('1', collection)
        self.assertIs(collection['1'], f1)

    def test_duplicate_ok(self):
        decor, collection = create_collector(
            raise_on_duplicate=False,
        )

        @decor('1')
        def f1():
            ...

        @decor('1')
        def f2():
            ...

        self.assertEqual(len(collection), 1)
        self.assertIn('1', collection)
        self.assertIs(collection['1'], f2)


class DecorClassTestCase(TestCase):
    def test_simple(self):
        decor, collection = create_class_collector()

        @decor
        class A:
            ...

        @decor
        class B:
            ...

        self.assertEqual(len(collection), 2)
        self.assertIn('A', collection)
        self.assertIs(collection['A'], A)
        self.assertIn('B', collection)
        self.assertIs(collection['B'], B)

    def test_key(self):
        decor, collection = create_class_collector()

        @decor('1')
        class A:
            ...

        @decor('2')
        class B:
            ...

        self.assertEqual(len(collection), 2)
        self.assertIn('1', collection)
        self.assertIs(collection['1'], A)
        self.assertIn('2', collection)
        self.assertIs(collection['2'], B)

    def test_duplicate_raise(self):
        decor, collection = create_class_collector(raise_on_duplicate=True)

        @decor
        class A:
            ...

        try:
            @decor
            class A:
                ...
        except ValueError as ex:
            self.assertTrue(str(ex).startswith('Duplication for key'))

        else:
            self.assertTrue(False, 'expected exception here')

        self.assertEqual(len(collection), 1)
        self.assertIn('A', collection)
        self.assertIs(collection['A'], A)

    def test_duplicate_ok(self):
        decor, collection = create_class_collector(raise_on_duplicate=False)

        @decor
        class A:
            ...

        @decor
        class A:
            ...

        self.assertEqual(len(collection), 1)
        self.assertIn('A', collection)
        self.assertIs(collection['A'], A)

    def test_key_getter(self):
        decor, collection = create_class_collector(key_getter=lambda x: x.x)

        @decor
        class A:
            x = 123

        @decor
        class B:
            x = 456

        self.assertEqual(len(collection), 2)
        self.assertIn(123, collection)
        self.assertIs(collection[123], A)
        self.assertIn(456, collection)
        self.assertIs(collection[456], B)




from typing import Callable, Type, TypeVar, Union, cast

KT = TypeVar('KT')
VT = TypeVar('VT')
WrapType = Callable[[VT], VT]
DecoratorType = Callable[[KT], WrapType[VT]]
CollectionType = dict[KT, VT]

ClassDecoratorType = Callable[[Union[KT, VT]], Union[VT, Callable[[VT], VT]]]
ClassCollectionType = dict[KT, VT]


def create_collector(
    raise_on_duplicate: bool = True,
) -> tuple[DecoratorType[KT, VT], CollectionType[KT, VT]]:
    def decor(key: KT) -> WrapType[VT]:
        def wrap(func: VT) -> VT:
            if key in collection and raise_on_duplicate:
                raise ValueError(f'Duplication for key {key}')
            collection[key] = func
            return func
        return wrap
    collection = {}  # type: CollectionType[KT, VT]
    return decor, collection


def create_class_collector(
    raise_on_duplicate: bool = True,
    key_getter: Callable[[Type[VT]], KT] = lambda x: x.__name__  # type: ignore[assignment,return-value]
) -> tuple[ClassDecoratorType[KT, Type[VT]], CollectionType[KT, Type[VT]]]:
    def decor(cls: Union[KT, Type[VT]]) -> Union[Type[VT], Callable[[Type[VT]], Type[VT]]]:

        def wrap(cls: Type[VT]) -> Type[VT]:
            if key in collection and raise_on_duplicate:
                raise ValueError(f'Duplication for key {key}')
            collection[key] = cls
            return cls

        if isinstance(cls, type):
            cls = cast(Type[VT], cls)
            key = key_getter(cls)  # type: KT
            return wrap(cls)

        key = cls
        return wrap

    collection = {}  # type: CollectionType[KT, Type[VT]]
    return decor, collection
